@@ -1,25 +1,31 @@
 # -*- coding: utf-8 -*-
 
 """
-Klient Google Readera
+Google Reader client.
 """
 
-# Mocno inspirowane (z kopiowaniem sporych fragmentów kodu włącznie):
+# Inspired by (including copying some code snippets):
 # http://blog.gpowered.net/2007/08/google-reader-api-functions.html
 #
-# Kluczowe informacje:
+# Key information I used initially:
 # http://code.google.com/p/pyrfeed/wiki/GoogleReaderAPI
+#
+# More docs, not yet fully consumed:
+# http://undoc.in/googlereader.html#search-items-ids
 
 import urllib
 import urllib2
 import re
-import simplejson
+import json
 import time
+from datetime import datetime
 from lxml import etree, objectify
 
 import logging
 log = logging.getLogger("reader")
 
+TRIM_LOG_MESSAGES_AT = 100
+
 TOKEN_VALID_TIME = 60
 #DUMP_REQUESTS = True
 #DUMP_REQUESTS = False
@@ -27,12 +33,18 @@ TOKEN_VALID_TIME = 60
 #DUMP_REPLIES = True
 
 class GoogleLoginFailed(Exception):
+    """
+    Exception raised on login failure and other authorization problems.
+    """
     pass
 class GoogleOperationFailed(Exception):
+    """
+    Exception raised when Google rejects some operation.
+    """
     pass
 
-#SOURCE = 'my-small-script'
-SOURCE = 'mekk'
+# User-agent/client-name
+SOURCE = 'mekk.reader_client'
 
 GOOGLE_URL = 'http://www.google.com'
 READER_URL = GOOGLE_URL + '/reader'
@@ -43,8 +55,12 @@ PREFERENCE_LIST_URL = READER_URL + '/api/0/preference/list'
 UNREAD_COUNT_URL = READER_URL + '/api/0/unread-count'
 SUBSCRIPTION_LIST_URL = READER_URL + '/api/0/subscription/list'
 SUBSCRIPTION_EDIT_URL = READER_URL + '/api/0/subscription/edit'
+SUBSCRIPTION_QUICKADD_URL = READER_URL + '/api/0/subscription/quickadd'
 TAG_EDIT_URL = READER_URL + '/api/0/edit-tag'
 TAG_DISABLE_URL = READER_URL + '/api/0/disable-tag'
+SEARCH_ITEMS_IDS_URL = READER_URL + '/api/0/search/items/ids'
+STREAM_ITEMS_CONTENTS_URL = READER_URL + '/api/0/stream/items/contents'
+STREAM_CONTENTS_FEED_URL = READER_URL + '/api/0/stream/contents/feed/%s'
 IN_STATE_URL = READER_URL + '/atom/user/-/state/com.google/%s'
 GET_FEED_URL = READER_URL + '/atom/feed/'
 READING_TAG_URL = READER_URL + '/atom/%s'
@@ -78,8 +94,8 @@ class GoogleReaderClient(object):
 
     def tag_id(self, tag):
         """
-        Converts tag name (say "Życie: Polityka" into 
-        tag id (say "user/-/label/Życie: Polityka").
+        Converts tag name (say "Life: Politics" into 
+        tag id (say "user/joe/label/Life: Politics").
         
         If parameter is already in this form, leaves it as-is
         """
@@ -88,6 +104,10 @@ class GoogleReaderClient(object):
         return tag
 
     def get_my_id(self):
+        """
+        Returns true user identifier to be used in API calls, calculating
+        it if necessary. Caches the result
+        """
         if self.my_id == '-':
             tl = self.get_tag_list()
             for vl in tl['tags']:
@@ -114,7 +134,24 @@ class GoogleReaderClient(object):
 
     def get_feed_atom(self, url, **kwargs):
         """
-        Atom feed for any feed. Works also for unsubscribed feeds
+        Atom feed for any feed. Works also for unsubscribed feeds.
+
+        Handled named parameters:
+
+        format: how should the reply be returned. Can be:
+            'xml' (raw xml text),
+            'etree' (lxml.etree object), or
+            'obj' (lxml.objectify object).
+          If not specified, 'obj' is default.
+
+        count: how many articles to get (default 20),
+
+        older_first: if specified and set to True, means returning older articles
+              first
+
+        continue_from: start from given article instead of the first one
+              (handle paging). Parameter given here should be taken from
+               <gr:continuation> value from the reply obtained earlier.
         """
         return self._get_atom(GET_FEED_URL + url,
                               **kwargs)
@@ -122,18 +159,27 @@ class GoogleReaderClient(object):
     def get_reading_list_atom(self, **kwargs):
         """
         Atom feed of unread items
+
+        Handles the same named parameters as get_feed_atom
+        (format, count, older_first, continue_from).
         """
         return self.get_instate_atom('reading-list', **kwargs)
 
     def get_read_atom(self, **kwargs):
         """
         Atom feed of (recent) read items
+
+        Handles the same named parameters as get_feed_atom
+        (format, count, older_first, continue_from).
         """
         return self.get_instate_atom('read', **kwargs)
 
     def get_tagged_atom(self, tag, **kwargs):
         """
         Atom feed of (unread?) items for given tag
+
+        Handles the same named parameters as get_feed_atom
+        (format, count, older_first, continue_from).
         """
         tagged_url = READING_TAG_URL % self.tag_id(tag)
         return self._get_atom(tagged_url, **kwargs)
@@ -141,18 +187,27 @@ class GoogleReaderClient(object):
     def get_starred_atom(self, **kwargs):
         """
         Atom feed of starred items
+
+        Handles the same named parameters as get_feed_atom
+        (format, count, older_first, continue_from).
         """
         return self.get_instate_atom('starred', **kwargs)
 
     def get_fresh_atom(self, **kwargs):
         """
         Atom feed of fresh (newly added) items
+
+        Handles the same named parameters as get_feed_atom
+        (format, count, older_first, continue_from).
         """
         return self.get_instate_atom('fresh', **kwargs)
 
     def get_broadcast_atom(self, **kwargs):
         """
         Atom feed of public (shared) items
+
+        Handles the same named parameters as get_feed_atom
+        (format, count, older_first, continue_from).
         """
         return self.get_instate_atom('broadcast', **kwargs)
 
@@ -165,8 +220,72 @@ class GoogleReaderClient(object):
         tracking-item-link-used, tracking-kept-unread
 
         get_fresh_atom is equivalent to get_instate_atom('fresh') and so on.
+
+        Handles the same named parameters as get_feed_atom
+        (format, count, older_first, continue_from).
         """
-        return self._get_atom(IN_STATE_URL % 'state', **kwargs)
+        return self._get_atom(IN_STATE_URL % state, **kwargs)
+
+    ############################################################
+    # Public API - item
+
+    def search_for_articles(self, query, count=20):
+        """
+        Searches for articles using given text query.
+        Returns plain list like:
+
+           [ u'7212740130471148824',
+             u'-8654279325215116158',
+             u'8121555931508499120',
+           ]
+
+        Those are short article(entry) identifiers
+        (note that reader also sometimes uses long form
+         'tag:google.com,2005:reader/item/5d0cfa30041d4348',
+         every reader api which requires article id
+         should handle both forms fine)
+        """
+        output = "json"
+        url = SEARCH_ITEMS_IDS_URL + "?"\
+              + urllib.urlencode({
+                "q": query,
+                "num": count,
+                "output": output,
+                "ck": int(time.mktime(datetime.now().timetuple())),
+                "client": SOURCE,
+                })
+        reply = json.loads(self._make_call(url))
+        return [ item['id'] for item in reply['results'] ]
+
+    def article_contents(self, ids):
+        """
+        Return article (entry) contents of specified articles. ids is
+        a list of identifiers (for example extracted from feed, or
+        returned from search_for_articles).
+
+        Returned structure is a complicated recursive dictionary
+        of which ['items'] list may be of biggest interest. Dump it for
+        details.
+        """
+        url = STREAM_ITEMS_CONTENTS_URL + "?" \
+              + urllib.urlencode({"ck": int(time.mktime(datetime.now().timetuple())),
+                                  "client": SOURCE})
+        post_params = [("i", id_) for id_ in ids]
+        #post_params.extend([("it", "0")] * len(post_params))
+        post_params.append(("T", self._get_token()))
+        return json.loads(self._make_call(url, post_params))
+
+    def feed_contents(self, feed_url, count=20, older_first=False):
+        """
+        Returns list of articles belonging to given feed.
+        """
+        url = STREAM_CONTENTS_FEED_URL % urllib.quote_plus(feed_url) + "?" \
+              + urllib.urlencode({
+                "ck": int(time.mktime(datetime.now().timetuple())),
+                "n": count,
+                "r": (older_first and "o" or "d"),
+                "client": SOURCE})
+        return json.loads(self._make_call(url))
 
 
     ############################################################
@@ -196,9 +315,45 @@ class GoogleReaderClient(object):
     ############################################################
     # Public API - subscription modifications
 
+    def subscribe_quickadd(self, site_url):
+        """
+        Subscribe to given site url.
+
+        Note: site_url is a normal address of website, Google Reader
+        will try to autodetect feed address.
+
+        Method returns a dictionary describing the call results.
+        In succesfull case it may look like:
+
+            {
+             u'numResults': 1,
+             u'query': u'http://sport.pl',
+             u'streamId': u'feed/http://rss.gazeta.pl/pub/rss/sport.xml'
+            }
+
+        In failed case (feed not found) it may look like:
+
+            {
+             u'numResults': 0,
+             u'query': u'http://sport.interia.pl',
+            }
+
+        """
+        url = SUBSCRIPTION_QUICKADD_URL + "?" \
+              + urllib.urlencode({"ck": int(time.mktime(datetime.now().timetuple())),
+                                  "client": SOURCE})
+        post_params = {
+            "quickadd": site_url,
+            "T": self._get_token(),
+            }
+        return json.loads(self._make_call(url, post_params))
+
     def subscribe_feed(self, feed_url, title = None):
         """
-        Subscribe to given feed. Optionally set title
+        Subscribe to given feed. Optionally set title.
+
+        Note: feed should specify RSS/Atom url. See subscribe_quickadd for
+        alternate method.
         """
         return self._change_feed(feed_url, 'subscribe', title = title)
 
@@ -238,7 +393,7 @@ class GoogleReaderClient(object):
         """
         Removes tag as a whole
         """
-        url = TAG_DISABLE_URL +  '?client=%s' % SOURCE
+        url = TAG_DISABLE_URL + '?client=%s' % SOURCE
         post_data = {
             's' : self.tag_id(tag),
             'ac' : 'disable-tags',
@@ -281,7 +436,7 @@ class GoogleReaderClient(object):
                 raise
         result = f.read()
 
-        log.debug("Result: %s" % result)
+        log.debug("Result: %s" % result[:TRIM_LOG_MESSAGES_AT])
 
         sid = re.search('Auth=(\S*)', result).group(1)
         if not sid:
@@ -299,10 +454,11 @@ class GoogleReaderClient(object):
             self.cached_token_time = t
         return self.cached_token
 
-    def _get_atom(self, url, count = None, older_first = False, continue_from = None, format = 'obj'):
+    def _get_atom(self, url, count = None, 
+                  older_first = False, continue_from = None, format = 'obj'):
         """
         Actually get ATOM feed. url is base url (one of the state or label urls).
-        count is the articles count (default 20), ordering_back set to False means older
+        count is the articles count (default 20), older_first set to True means older
         first, continue_from can be set to gr:continuation value from the feed to
         grab more items
 
@@ -326,7 +482,8 @@ class GoogleReaderClient(object):
         else:
             return r
 
-    def _change_feed(self, feed_url, operation, title = None, add_tag = None, remove_tag = None):
+    def _change_feed(self, feed_url, operation,
+                     title = None, add_tag = None, remove_tag = None):
         """
         Subscribe or unsubscribe
         """
@@ -383,17 +540,32 @@ class GoogleReaderClient(object):
 
     def _get_list(self, url, format):
         if format == 'obj':
-            return simplejson.loads(
+            return json.loads(
                 self._make_call(url + '?output=json'))
         else:
             return self._make_call(url + '?output=' + format)
         
 
     def _make_call(self, url, post_data=None):
+        """
+        Actually executes a call to given url, adding authorization headers
+        and parameters.
+        
+        post_data can be either a dictionary, or list of (key, value)
+        pairs. In both cases value should be unicode (and will be encoded
+        to utf-8 inside this method).
+        """
         header = {'User-agent' : SOURCE}
         header['Authorization'] = 'GoogleLogin auth=%s' % self.session_id
         if post_data is not None:
-            true_data = [ (key, value.encode('utf-8')) for key, value in post_data.iteritems() ]
+            if type(post_data) is list:
+                true_data = [
+                    (key, value.encode('utf-8')) 
+                    for key, value in post_data ]
+            else:
+                true_data = [
+                    (key, value.encode('utf-8')) 
+                    for key, value in post_data.iteritems() ]
             true_data = urllib.urlencode(true_data)
         else:
             true_data = None
@@ -402,14 +574,14 @@ class GoogleReaderClient(object):
         if log.isEnabledFor("info"):
             if post_data:
                 log.info("Calling %s with parameters:\n    %s" % (
-                        request.get_full_url(), str(post_data)))
+                        request.get_full_url(), str(post_data)[:TRIM_LOG_MESSAGES_AT]))
             else:
                 log.info("Calling %s" % request.get_full_url())
 
         f = urllib2.urlopen( request )
         result = f.read()
 
-        log.debug("Result: %s" % result)
+        log.debug("Result: %s" % result[:TRIM_LOG_MESSAGES_AT])
 
         return result
 
